#include <curl/curl.h>
#include <storage/file.h>

#include "queuenode.h"
#include "server.h"

QueueNode::QueueNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nType, int nID )
{
	m_pcServer = pcServer;
	
	m_zLocalPath = zLocalPath;
	m_zRemotePath = zRemotePath;
	m_nType = nType;
	m_nStatus = NODE_QUEUED;
	m_nID = nID;
}

/* Set the necessary options in the curl handle. */
/* The important parts are done in the subclasses. */
CURLcode QueueNode::SetupCurlHandleForTransfer( CURL* pHandle )
{
	/* Call curl_easy_setopt() to set url, callbacks, etc. */
	/* We set some general options here; specifics like resume position etc will be set by subclasses. */

	/* Set callback functions */
	curl_easy_setopt( pHandle, CURLOPT_WRITEFUNCTION, QueueNode::WriteCallback );
	curl_easy_setopt( pHandle, CURLOPT_WRITEDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_READFUNCTION, QueueNode::ReadCallback );
	curl_easy_setopt( pHandle, CURLOPT_READDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_SEEKFUNCTION, QueueNode::SeekCallback );
	curl_easy_setopt( pHandle, CURLOPT_SEEKDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_PROGRESSFUNCTION, QueueNode::ProgressBarCallback );
	curl_easy_setopt( pHandle, CURLOPT_PROGRESSDATA, this );
	curl_easy_setopt( pHandle, CURLOPT_NOPROGRESS, 0 );
	
	return( CURLE_OK );
}

QueueNode::~QueueNode()
{
}


/** The libcurl callback functions. Can't pass a virtual function to libcurl so we use these static wrappers. **/
size_t QueueNode::ReadCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->Read( pBuf, nSize * nMult ) );
}

size_t QueueNode::WriteCallback( void* pBuf, size_t nSize, size_t nMult, void* pCookie )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->Write( pBuf, nSize * nMult ) );
}

int QueueNode::SeekCallback( void* pCookie, curl_off_t nOffset, int nOrigin )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->Seek( nOffset, nOrigin ) );
}

int QueueNode::ProgressBarCallback( void* pCookie, double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	QueueNode* pcTransfer = (QueueNode*)pCookie;
	return( pcTransfer->ProgressBar( fDownTotal, fDownNow, fUpTotal, fUpNow ) );
}

/**** DownloadNode class ****/
DownloadNode::DownloadNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID )
	: QueueNode( pcServer, zLocalPath, zRemotePath, NODE_DOWNLOAD, nID )
{
	m_pcFile = NULL;	/* This is only created when SetupCurlHandleForTransfer() is first called. */
}

DownloadNode::~DownloadNode()
{
	if( m_pcFile ) delete( m_pcFile );
}

CURLcode DownloadNode::SetupCurlHandleForTransfer( CURL* pHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	if( m_pcFile == NULL ) {
		m_pcFile = new File( m_zLocalPath, O_WRONLY | O_CREAT | O_TRUNC );
	}
	
	QueueNode::SetupCurlHandleForTransfer( pHandle );
	

	curl_off_t nOffset = m_pcFile->Seek( 0, SEEK_CUR );	/* Get current position in file, for if we are resuming */
	
	/* Set the url, set to download */
	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";
	zURL += m_zRemotePath;
	printf( "DownloadNode: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_UPLOAD, 0 );
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, nOffset );
	
	return( CURLE_OK );
}

size_t DownloadNode::Write( void* pBuf, size_t nSize )
{
/*	char zTmp[1025];
	memcpy( zTmp, pBuf, 1024 );
	zTmp[1024] = 0;
	printf( "DownloadNode::Write( %i bytes ):\n%s\n", nSize, zTmp );
*/
//	printf( "DownloadNode::Write( %i bytes ) to %s\n", nSize, m_zLocalPath.c_str() );
	size_t nResult = m_pcFile->Write( pBuf, nSize );
	m_pcFile->Flush();
	return( nResult );
}

int DownloadNode::Seek( curl_off_t nOffset, int nOrigin )
{
	return( m_pcFile->Seek( nOffset, nOrigin ) );
	/* TODO: make sure that File::Seek() returns the same as libcurl expects */
}

int DownloadNode::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	printf( "Progress for download %s: %.1f%%\n", m_zRemotePath.c_str(), fDownNow/fDownTotal * 100 );
	return( 0 );
}

/**** UploadNode class ****/
UploadNode::UploadNode( Server* pcServer, const String& zLocalPath, const String& zRemotePath, int nID )
	: QueueNode( pcServer, zLocalPath, zRemotePath, NODE_UPLOAD, nID )
{
	m_pcFile = NULL;	/* This is only created when SetupCurlHandleForTransfer() is first called. */
}

UploadNode::~UploadNode()
{
	if( m_pcFile ) delete( m_pcFile );
}

CURLcode UploadNode::SetupCurlHandleForTransfer( CURL* pHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	if( m_pcFile == NULL ) {
		m_pcFile = new File( m_zLocalPath, O_RDONLY );
	}
	
	QueueNode::SetupCurlHandleForTransfer( pHandle );
	

#if 1
	/* I don't think resuming uploads is this easy with ilbcurl :( */
	curl_off_t nOffset = m_pcFile->Seek( 0, SEEK_CUR );	/* Get current position in file, for if we are resuming */
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, nOffset );
#else
	m_pcFile->Seek( 0, SEEK_SET );
	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
#endif

	curl_off_t nSize = m_pcFile->GetSize();
	curl_easy_setopt( pHandle, CURLOPT_INFILESIZE_LARGE, nSize );
	
	/* Set the url, set to upload */
	String zURL = m_pcServer->GetServerAddress();
	zURL += "/";
	zURL += m_zRemotePath;
	printf( "UploadNode: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_UPLOAD, 1 );
	
	return( CURLE_OK );
}

size_t UploadNode::Read( void* pBuf, size_t nSize )
{
//	printf( "UploadNode::Read( %i )\n", nSize );
	return( m_pcFile->Read( pBuf, nSize ) );
}

int UploadNode::Seek( curl_off_t nOffset, int nOrigin )
{
	return( m_pcFile->Seek( nOffset, nOrigin ) );
	/* TODO: make sure that File::Seek() returns the same as libcurl expects */
}

int UploadNode::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
	printf( "Progress for upload %s: %.1f%%\n", m_zRemotePath.c_str(), (float)(100 * fUpNow/fUpTotal) );
	return( 0 );
}

/**** DirListNode class ****/
DirListNode::DirListNode( Server* pcServer, const String& zRemotePath, Handler* pcTarget, int nID )
	: QueueNode( pcServer, "", zRemotePath, NODE_DIRLIST, nID )
{
	m_pcTarget = pcTarget;
	m_bInitial = true;
}

DirListNode::~DirListNode()
{
}

CURLcode DirListNode::SetupCurlHandleForTransfer( CURL* pHandle )
{
	String zServer = m_pcServer->GetServerAddress(); 	/* Address of the form "ftp://user:pwd@host:port" (no path) */
	
	QueueNode::SetupCurlHandleForTransfer( pHandle );
	
	String zURL = zServer;
	zURL += "/";
	zURL += m_zRemotePath;
	zURL += "/";			/* Make it clear to libcurl that we want a dir listing */
	

	/* Set the url, set to upload */
	printf( "DirListNode: setting url to %s\n", zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_URL, zURL.c_str() );
	curl_easy_setopt( pHandle, CURLOPT_UPLOAD, 0 );
//	curl_easy_setopt( pHandle, CURLOPT_RESUME_FROM_LARGE, (curl_off_t)0 );
	
	return( CURLE_OK );
}

size_t DirListNode::Write( void* pBuf, size_t nSize )
{
	size_t nResult = m_pcServer->ParseDirListing( m_zRemotePath, (char*)pBuf, nSize, m_pcTarget, m_bInitial );
	m_bInitial = false;
	return( nResult );
}

int DirListNode::ProgressBar( double fDownTotal, double fDownNow, double fUpTotal, double fUpNow )
{
//	printf( "Progress for dirlisting %s: %.1f out of %.1f\n", m_zRemotePath.c_str(), (float)fDownNow, (float)fDownTotal );
	return( 0 );
}


